/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "houdini/houdini.pmh"

namespace fe
{

void SurfaceAccessibleHoudini::append(
	sp<SurfaceAccessibleI> a_spSurfaceAccessibleI,
	SpatialTransform& a_rTransform)
{
	SAFEGUARD;

	feLog("SurfaceAccessibleHoudini::append\n");
	sp<SurfaceAccessibleHoudini> spSurfaceOther=
			a_spSurfaceAccessibleI;
	if(!spSurfaceOther.isValid())
	{
		//* TODO support cross copies
		feLog("SurfaceAccessibleHoudini::append"
				" not SurfaceAccessibleHoudini\n");
		return;
	}

	UT_Matrix4 utMatrix4(
			a_rTransform(0,0),a_rTransform(1,0),a_rTransform(2,0),0.0,
			a_rTransform(0,1),a_rTransform(1,1),a_rTransform(2,1),0.0,
			a_rTransform(0,2),a_rTransform(1,2),a_rTransform(2,2),0.0,
			a_rTransform(0,3),a_rTransform(1,3),a_rTransform(2,3),1.0);

	const GU_Detail* pOtherGdp=spSurfaceOther->gdp();
	GU_Detail copyGdp;
	copyGdp.duplicate(*pOtherGdp);
	copyGdp.transform(utMatrix4);
	const_cast<GU_Detail*>(m_pGdp)->merge(copyGdp);

	m_primitiveMap.clear();
}

void SurfaceAccessibleHoudini::instance(sp<SurfaceAccessibleI>
	a_spSurfaceAccessibleI,std::vector<SpatialTransform>& a_rTransformArray)
{
	//* TODO
}

void SurfaceAccessibleHoudini::initPrimitiveMap(void)
{
	if(!m_pGdp)
	{
		return;
	}

	m_hasDiscretePolygons=FALSE;

#ifdef FE_HOUDINI_USE_GA
	const I32 primitiveCount=m_pGdp->getNumPrimitives();
#else
	const I32 primitiveCount=m_pGdp->primitives().entries();
#endif

	for(I32 index=0;index<primitiveCount;index++)
	{
#ifdef FE_HOUDINI_USE_GA
		const GEO_Primitive* pPrimitive=
				m_pGdp->getGEOPrimitive(m_pGdp->primitiveOffset(index));
#else
		const GEO_Primitive* pPrimitive=m_pGdp->primitives()(index);
#endif

#ifdef FE_HOUDINI_USE_GA
		if(pPrimitive->getTypeId().get()==GEO_PRIMPOLY)
#else
		if(pPrimitive->getPrimitiveId()==GEOPRIMPOLY)
#endif
		{
			m_hasDiscretePolygons=TRUE;
			break;
		}
	}

	I32 mapCount=0;
	if(m_hasDiscretePolygons)
	{
		m_primitiveMap[mapCount++]= -1;
	}

	for(I32 index=0;index<primitiveCount;index++)
	{
#ifdef FE_HOUDINI_USE_GA
		const GEO_Primitive* pPrimitive=
				m_pGdp->getGEOPrimitive(m_pGdp->primitiveOffset(index));
#else
		const GEO_Primitive* pPrimitive=m_pGdp->primitives()(index);
#endif

#ifdef FE_HOUDINI_USE_GA
		if(pPrimitive->getTypeId().get()==GEO_PRIMPOLY)
#else
		if(pPrimitive->getPrimitiveId()==GEOPRIMPOLY)
#endif
		{
			continue;
		}

		m_primitiveMap[mapCount++]=index;
	}
}

I32 SurfaceAccessibleHoudini::lookupPrimitive(I32 a_surfaceIndex) const
{
	std::map<I32,I32>::const_iterator it=m_primitiveMap.find(a_surfaceIndex);
	if(it==m_primitiveMap.end())
	{
		return -2;
	}
	return it->second;
}

//* TODO SurfaceBag
sp<SurfaceI> SurfaceAccessibleHoudini::surface(void)
{
	return surface(String());
}

sp<SurfaceI> SurfaceAccessibleHoudini::surface(String a_group)
{
	if(!m_pGdp)
	{
		return sp<SurfaceI>(NULL);
	}

	SAFEGUARD;
	if(m_spSurfaceI.isNull())
	{
#ifdef FE_HOUDINI_USE_GA
		const I32 entries=m_pGdp->getNumPrimitives();
#else
		const I32 entries=m_pGdp->primitives().entries();
#endif

		if(entries==1)
		{
			//* single primitive (perhaps a sphere)
			return subSurface(0,a_group);
		}

#ifdef FE_HOUDINI_USE_GA
		const GEO_Primitive* pPrimitive=
				m_pGdp->getGEOPrimitive(m_pGdp->primitiveOffset(0));
#else
		const GEO_Primitive* pPrimitive=m_pGdp->primitives()(0);
#endif

		//* HACK presuming they're all the same
		const GEO_Face* pGeoFace=dynamic_cast<const GEO_Face*>(pPrimitive);
		const BWORD openPoly=pGeoFace && !pGeoFace->isClosed();
		if(openPoly)
		{
			//* HACK presuming uniform CV counts
			const U32 subCount=pGeoFace->getVertexCount();

			//* all curves
			m_spSurfaceI=registry()->create("SurfaceI.SurfaceCurvesHoudini");
			if(m_spSurfaceI.isValid())
			{
				sp<SurfaceCurvesHoudini> spSurface=m_spSurfaceI;
				if(spSurface.isValid())
				{
					spSurface->setGdp(m_pGdp);
					spSurface->setGroup(a_group);
					spSurface->setGrain(subCount);
					spSurface->setOpenCurve(TRUE);
				}
			}
		}

		//* all triangles (presumed)
		m_spSurfaceI=registry()->create("SurfaceI.SurfaceTrianglesHoudini");
		if(m_spSurfaceI.isValid())
		{
			sp<SurfaceTrianglesHoudini> spSurface=m_spSurfaceI;
			if(spSurface.isValid())
			{
				spSurface->setGdp(m_pGdp);
				spSurface->setGroup(a_group);
			}
		}
	}

	return m_spSurfaceI;
}

sp<SurfaceI> SurfaceAccessibleHoudini::subSurface(U32 a_subIndex)
{
	return subSurface(a_subIndex,String());
}

sp<SurfaceI> SurfaceAccessibleHoudini::subSurface(U32 a_subIndex,String a_group)
{
	if(!m_pGdp)
	{
		return sp<SurfaceI>(NULL);
	}

	SAFEGUARD;

	if(m_primitiveMap.empty())
	{
		initPrimitiveMap();
	}

	const I32 primitiveIndex=lookupPrimitive(a_subIndex);
	if(primitiveIndex< -1)
	{
		return sp<SurfaceI>(NULL);
	}

	if(primitiveIndex>=0)
	{
#ifdef FE_HOUDINI_USE_GA
		const GEO_Primitive* pPrimitive=m_pGdp->getGEOPrimitive(
				m_pGdp->primitiveOffset(primitiveIndex));
#else
		const GEO_Primitive* pPrimitive=m_pGdp->primitives()(primitiveIndex);
#endif

		if(!pPrimitive)
		{
			return sp<SurfaceI>(NULL);
		}

#ifdef FE_HOUDINI_USE_GA
		if(pPrimitive->getTypeId().get()==GEO_PRIMSPHERE)
#else
		if(pPrimitive->getPrimitiveId()==GEOPRIMSPHERE)
#endif
		{
			sp<SurfaceSphere> spSphere=
					registry()->create("SurfaceI.SurfaceSphere");
			if(!spSphere.isValid())
			{
				return sp<SurfaceI>(NULL);
			}

			const GEO_PrimSphere* pPrimSphere=
					dynamic_cast<const GEO_PrimSphere*>(pPrimitive);

			UT_Matrix4 utMatrix4;
			pPrimSphere->getTransform4(utMatrix4);
			SpatialTransform matrix;
			set(matrix,utMatrix4.data());

//			fpointOffseteLog("SurfaceAccessibleHoudini::subSurface(%d) prim %d"
//					" sphere matrix\n%s\n",
//					a_subIndex,primitiveIndex,c_print(matrix));

			spSphere->setCenter(matrix.translation());
			spSphere->setRadius(matrix(0,0));

			return spSphere;
		}

#ifdef FE_HOUDINI_USE_GA
		if(pPrimitive->getTypeId().get()==GEO_PRIMCIRCLE)
#else
		if(pPrimitive->getPrimitiveId()==GEOPRIMCIRCLE)
#endif
		{
			sp<SurfaceDisk> spDisk=
					registry()->create("SurfaceI.SurfaceDisk");
			if(!spDisk.isValid())
			{
				return sp<SurfaceI>(NULL);
			}

			const GEO_PrimCircle* pPrimCircle=
					dynamic_cast<const GEO_PrimCircle*>(pPrimitive);

			UT_Matrix4 utMatrix4;
			pPrimCircle->getTransform4(utMatrix4);
			SpatialTransform matrix;
			set(matrix,utMatrix4.data());

			UT_Vector3 utVector3=pPrimCircle->computeNormal();

//			feLog("SurfaceAccessibleHoudini::subSurface(%d) prim %d"
//					" disk matrix\n%s\n",
//					a_subIndex,primitiveIndex,c_print(matrix));

			spDisk->setCenter(matrix.translation());
			spDisk->setRadius(matrix(0,0));
			spDisk->setSpan(
					SpatialVector(utVector3[0],utVector3[1],utVector3[2]));

			return spDisk;
		}

#ifdef FE_HOUDINI_USE_GA
		if(pPrimitive->getTypeId().get()==GEO_PRIMTUBE)
#else
		if(pPrimitive->getPrimitiveId()==GEOPRIMTUBE)
#endif
		{
			sp<SurfaceCylinder> spCylinder=
					registry()->create("SurfaceI.SurfaceCylinder");
			if(!spCylinder.isValid())
			{
				return sp<SurfaceI>(NULL);
			}

			const GEO_PrimTube* pPrimTube=
					dynamic_cast<const GEO_PrimTube*>(pPrimitive);

			UT_Matrix4 utMatrix4;
			pPrimTube->getTransform4(utMatrix4);
			SpatialTransform matrix;
			set(matrix,utMatrix4.data());

			const Real taper=pPrimTube->getTaper();

//			feLog("SurfaceAccessibleHoudini::subSurface(%d) prim %d"
//					" cylinder taper %.6G matrix\n%s\n",
//					a_subIndex,primitiveIndex,taper,c_print(matrix));

			const Real scaleX=magnitude(matrix.column(0)); // same as scaleZ

//			feLog("scale %.6G %.6G %.6G\n",scaleX,scaleY,scaleZ);

			const SpatialVector span=matrix.column(1);	//* already scaled

			spCylinder->setLocation(matrix.translation()-0.5*span);
			spCylinder->setSpan(span);
			spCylinder->setBaseRadius(scaleX);
			spCylinder->setEndRadius(scaleX*taper);

			return spCylinder;
		}
	}

	sp<SurfaceTrianglesHoudini> spTriangles=
			registry()->create("SurfaceI.SurfaceTrianglesHoudini");
	if(!spTriangles.isValid())
	{
		return sp<SurfaceI>(NULL);
	}

	spTriangles->setGdp(m_pGdp);
	spTriangles->setGroup(a_group);
	if(primitiveIndex>=0)
	{
		spTriangles->setSubIndex(a_subIndex);
	}

	return spTriangles;
}

} // namespace
