/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __surface_SurfaceSearchable_h__
#define __surface_SurfaceSearchable_h__

namespace fe
{

/**************************************************************************//**
	@brief Tree-searchable Surface

	@ingroup surface
*//***************************************************************************/
class FE_DL_EXPORT SurfaceSearchable: public SurfaceSphere
{
	public:

	class Impact: public SurfaceSphere::Impact
	{
		public:
							Impact(void)
							{
#if FE_COUNTED_STORE_TRACKER
								setName("SurfaceSearchable::Impact");
#endif
							}
	virtual					~Impact(void)									{}

	virtual	SpatialVector	normal(void) const
							{	return m_normal; }
			void			setNormal(SpatialVector a_normal)
							{	m_normal=a_normal; }

	virtual	I32				pickIndex(void) const
							{	return m_pickIndex; }
			void			setPickIndex(I32 a_pickIndex)
							{	m_pickIndex=a_pickIndex; }


	virtual	void			copy(sp<SpatialTreeI::Hit>& a_rspHit);

		protected:
			SpatialVector	m_normal;
			I32				m_pickIndex;
	};

							SurfaceSearchable(void);
virtual						~SurfaceSearchable(void);

							//* As Protectable
virtual	Protectable*		clone(Protectable* pInstance=NULL);

							//* As SurfaceI
virtual	void				setRefinement(U32 a_refinement)
							{
								m_refinement=a_refinement;
								if(m_spSpatialTreeI.isValid())
								{
									m_spSpatialTreeI->setRefinement(
											m_refinement);
								}
							}

virtual SpatialTransform	sample(Vector2 a_uv) const;

virtual	void				prepareForSearch(void);

virtual	std::vector< sp<SurfaceI::ImpactI> >
							nearestPoints(const SpatialVector& a_origin,
									U32 a_hitLimit) const;
virtual	sp<ImpactI>			nearestPoint(const SpatialVector& a_origin) const;

virtual	std::vector< sp<SurfaceI::ImpactI> >
							rayImpacts(const SpatialVector& a_origin,
									const SpatialVector& a_direction,
									Real a_maxDistance,U32 a_hitLimit) const;
virtual	sp<ImpactI>			rayImpact(const SpatialVector& a_origin,
									const SpatialVector& a_direction,
									Real a_maxDistance) const;

	protected:

virtual	void				cache(void);

		void				calcBoundingSphere(void);
		void				resizeArrays(U32 a_count);

		U32					m_refinement;
		U32					m_vertices;
		SpatialVector*		m_pVertexArray;
		SpatialVector*		m_pNormalArray;
		I32*				m_pPickIndexArray;
		sp<Component>		m_spSearchComponent;
		sp<SpatialTreeI>	m_spSpatialTreeI;

	protected:
mutable	CountedPoolCore*	m_pImpactPool;
};

} // namespace

#endif /* __surface_SurfaceSearchable_h__ */
